"""
The module provides help methods to calculate size of nontrivial types.
"""

import typing

from zserio.bitbuffer import BitBuffer
from zserio.exception import PythonRuntimeException
from zserio.limits import INT64_MIN

def getBitSizeOfVarInt16(value: int) -> int:
    """
    Gets bit size of variable 16-bit signed integer value.

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varint16 type.
    """

    return _getBitSizeOfVarIntImpl(abs(value), VARINT16_MAX_VALUES, "varint16")

def getBitSizeOfVarInt32(value: int) -> int:
    """
    Gets bit size of variable 32-bit signed integer value.

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varint32 type.
    """

    return _getBitSizeOfVarIntImpl(abs(value), VARINT32_MAX_VALUES, "varint32")

def getBitSizeOfVarInt64(value: int) -> int:
    """
    Gets bit size of variable 64-bit signed integer value.

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varint64 type.
    """

    return _getBitSizeOfVarIntImpl(abs(value), VARINT64_MAX_VALUES, "varint64")

def getBitSizeOfVarInt(value: int) -> int:
    """
    Gets bit size of variable signed integer value (up to 9 bytes).

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varint type.
    """

    if value == INT64_MIN:
        return 8 # INT64_MIN is stored as -0
    return _getBitSizeOfVarIntImpl(abs(value), VARINT_MAX_VALUES, "varint")

def getBitSizeOfVarUInt16(value: int) -> int:
    """
    Gets bit size of variable 16-bit unsigned integer value.

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varuint16 type.
    """

    return _getBitSizeOfVarIntImpl(value, VARUINT16_MAX_VALUES, "varuint16")

def getBitSizeOfVarUInt32(value: int) -> int:
    """
    Gets bit size of variable 32-bit unsigned integer value.

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varuint32 type.
    """

    return _getBitSizeOfVarIntImpl(value, VARUINT32_MAX_VALUES, "varuint32")

def getBitSizeOfVarUInt64(value: int) -> int:
    """
    Gets bit size of variable 64-bit unsigned integer value.

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varuint64 type.
    """

    return _getBitSizeOfVarIntImpl(value, VARUINT64_MAX_VALUES, "varuint64")

def getBitSizeOfVarUInt(value: int) -> int:
    """
    Gets bit size of variable unsigned integer value (up to 9 bytes).

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varuint type.
    """

    return _getBitSizeOfVarIntImpl(value, VARUINT_MAX_VALUES, "varuint")

def getBitSizeOfVarSize(value: int) -> int:
    """
    Gets bit size of variable size integer value.

    :param value: Value to use for bit size calculation.
    :returns: Bit size of the value.
    :raises PythonRuntimeException: Throws if given value is out of range for varsize type.
    """

    return _getBitSizeOfVarIntImpl(value, VARSIZE_MAX_VALUES, "varsize")

def getBitSizeOfString(string: str) -> int:
    """
    Gets bit size of string.

    :param string: String value to use for bit size calculation.
    :raises PythonRuntimeException: Throws if given string is too long.
    """

    stringBytes = string.encode("utf-8")
    return getBitSizeOfVarSize(len(stringBytes)) + len(stringBytes) * 8

def getBitSizeOfBitBuffer(bitBuffer: BitBuffer) -> int:
    """
    Gets the bit size of bit buffer which is stored in bit stream.

    :param bitBuffer: Bit buffer for calculation.
    :returns: Length of bit buffer in bits.
    :raises PythonRuntimeException: Throws if given bit buffer is too long.
    """
    bitBufferSize = bitBuffer.getBitSize()

    # bit buffer consists of varsize for bit size followed by the bits
    return getBitSizeOfVarSize(bitBufferSize) + bitBufferSize

def _getBitSizeOfVarIntImpl(value: int, maxValues: typing.Sequence[int], varIntName: str) -> int:
    if value >= 0:
        absValue = abs(value)
        for i, maxValue in enumerate(maxValues):
            if absValue <= maxValue:
                return (i + 1) * 8

    raise PythonRuntimeException("bitsizeof: Value '%d' is out of range for %s!" % (value, varIntName))

VARINT16_MAX_VALUES = [
    (1 << (6)) - 1,
    (1 << (6 + 8)) - 1
]

VARINT32_MAX_VALUES = [
    (1 << (6)) - 1,
    (1 << (6 + 7)) - 1,
    (1 << (6 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 8)) - 1,
]

VARINT64_MAX_VALUES = [
    (1 << (6)) - 1,
    (1 << (6 + 7)) - 1,
    (1 << (6 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7 + 7 + 7 + 8)) - 1,
]

VARINT_MAX_VALUES = [
    (1 << (6)) - 1,
    (1 << (6 + 7)) - 1,
    (1 << (6 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (6 + 7 + 7 + 7 + 7 + 7 + 7 + 7 + 8)) - 1
]

VARUINT16_MAX_VALUES = [
    (1 << (7)) - 1,
    (1 << (7 + 8)) - 1
]

VARUINT32_MAX_VALUES = [
    (1 << (7)) - 1,
    (1 << (7 + 7)) - 1,
    (1 << (7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 8)) - 1
]

VARUINT64_MAX_VALUES = [
    (1 << (7)) - 1,
    (1 << (7 + 7)) - 1,
    (1 << (7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7 + 7 + 7 + 8)) - 1
]

VARUINT_MAX_VALUES = [
    (1 << (7)) - 1,
    (1 << (7 + 7)) - 1,
    (1 << (7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7 + 7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7 + 7 + 7 + 7 + 7 + 8)) - 1,
]

VARSIZE_MAX_VALUES = [
    (1 << (7)) - 1,
    (1 << (7 + 7)) - 1,
    (1 << (7 + 7 + 7)) - 1,
    (1 << (7 + 7 + 7 + 7)) - 1,
    (1 << (2 + 7 + 7 + 7 + 8)) - 1,
]



"""
The module provides help methods for bit fields calculation.
"""

from zserio.exception import PythonRuntimeException

def getBitFieldLowerBound(length: int) -> int:
    """
    Gets the lower bound of a unsigned bitfield type with given length.

    :param length: Length of the unsigned bitfield in bits.
    :returns: The lowest value the unsigned bitfield can hold.
    :raises PythonRuntimeException: If unsigned bitfield with wrong length has been specified.
    """

    _checkBitFieldLength(length)
    return 0

def getBitFieldUpperBound(length: int) -> int:

    """
    Gets the upper bound of a unsigned bitfield type with given length.

    :param length: Length of the unsigned bitfield in bits.
    :returns: The largest value the unsigned bitfield can hold.
    :raises PythonRuntimeException: If unsigned bitfield with wrong length has been specified.
    """

    _checkBitFieldLength(length)
    return (1 << length) - 1

def getSignedBitFieldLowerBound(length: int) -> int:
    """
    Gets the lower bound of a signed bitfield type with given length.

    :param length: Length of the signed bitfield in bits.
    :returns: The lowest value the signed bitfield can hold.
    :raises PythonRuntimeException: If signed bitfield with wrong length has been specified.
    """

    _checkBitFieldLength(length)
    return -(1 << (length - 1))

def getSignedBitFieldUpperBound(length: int) -> int:
    """
    Gets the upper bound of a signed bitfield type with given length.

    :param length: Length of the signed bitfield in bits.
    :returns: The largest value the signed bitfield can hold.
    :raises PythonRuntimeException: If signed bitfield with wrong length has been specified.
    """

    _checkBitFieldLength(length)
    return (1 << (length - 1)) - 1

def _checkBitFieldLength(length: int) -> None:
    if length <= 0 or length > MAX_BITFIELD_BITS:
        raise PythonRuntimeException("bitfield: Asking for bound of bitfield with invalid length %d!" % length)

MAX_BITFIELD_BITS = 64



"""
The module provides classes for Zserio Pub/Sub.
"""

import typing

from zserio.exception import PythonRuntimeException

class PubsubInterface:
    """ Interface for Pub/Sub client backends. """

    def publish(self, topic: str, data: bytes, context: typing.Any = None) -> None:
        """
        Publishes given data as a specified topic.

        :param topic: Topic definition.
        :param data: Data to publish.
        :param context: Context specific for a particular Pub/Sub implementation.
        :raises PubsubException: If the publishing fails.
        """
        raise NotImplementedError()

    def subscribe(self, topic: str, callback: typing.Callable[[str, bytes], None],
                  context: typing.Any = None) -> int:
        """
        Subscribes a topic.

        :param topic: Topic definition to subscribe. Note that the definition format depends on the particular
                     Pub/Sub backend implementation and therefore e.g. wildcards can be used only if they are
                     supported by Pub/Sub backend.
        :param callback: Callback to be called when a message with the specified topic arrives.
        :param context: Context specific for a particular Pub/Sub implementation.
        :returns: Subscription ID.
        :raises PubsubException: If subscribing fails.
        """
        raise NotImplementedError()

    def unsubscribe(self, subscriptionId: int) -> None:
        """
        Unsubscribes the subscription with the given ID.

        :param id: ID of the subscription to be unsubscribed.
        :raises PubsubException: If unsubscribing fails.
        """
        raise NotImplementedError()

class PubsubException(PythonRuntimeException):
    """
    Exception thrown in case of an error in Zserio Pub/Sub.
    """



"""
The module implements abstraction for reading data to the bit stream.
"""

import typing

from zserio.bitbuffer import BitBuffer
from zserio.bitsizeof import INT64_MIN
from zserio.exception import PythonRuntimeException
from zserio.float import convertUInt16ToFloat, convertUInt32ToFloat, convertUInt64ToFloat

class BitStreamReader:
    """
    Bit stream reader.
    """

    def __init__(self, buffer: bytes) -> None:
        """
        Constructs bit stream reader from bytes buffer.

        :param buffer: Bytes-like buffer to read as a bit stream.
        """

        self._buffer = buffer # type: bytes
        self._bitPosition = 0 # type: int
        self._bitSize = len(buffer) * 8 # type: int

    @classmethod
    def fromBitBuffer(cls: typing.Type['BitStreamReader'], bitBuffer: BitBuffer) -> 'BitStreamReader':
        """
        Constructs bit stream reader from bit buffer.

        :param bitBuffer: Bit buffer to read as a bit stream.
        """

        instance = cls(bitBuffer.getBuffer())
        instance._bitSize = bitBuffer.getBitSize()

        return instance

    @classmethod
    def fromFile(cls: typing.Type['BitStreamReader'], filename: str) -> 'BitStreamReader':
        """
        Constructs bit stream reader from file.

        :param filename: Filename to read as a bit stream.
        """

        with open(filename, 'rb') as file:
            return cls(file.read())

    def readBits(self, numBits: int) -> int:
        """
        Reads given number of bits from the bit stream as an unsigned integer.

        :param numBits: Number of bits to read.
        :returns: Read bits as an unsigned integer.
        :raises PythonRuntimeException: If the numBits is invalid number of the reading goes behind the stream.
        """

        if numBits < 0:
            raise PythonRuntimeException("BitStreamReader: Reading negative number of bits!")

        endBitPosition = self._bitPosition + numBits

        if endBitPosition > self._bitSize:
            raise PythonRuntimeException("BitStreamReader: Reading behind the stream!")

        startByte = self._bitPosition // 8
        endByte = (endBitPosition - 1) // 8

        value = int.from_bytes(self._buffer[startByte : endByte + 1], byteorder='big', signed=False)

        lastBits = endBitPosition % 8
        if lastBits != 0:
            value >>= (8 - lastBits)
        value &= (1 << numBits) - 1

        self._bitPosition = endBitPosition

        return value

    def readSignedBits(self, numBits: int) -> int:
        """
        Reads given number of bits from the bit stream as a signed integer.

        :param numBits: Number of bits to read
        :returns: Read bits as a signed integer.
        :raises PythonRuntimeException: If the numBits is invalid number of the reading goes behind the stream.
        """

        value = self.readBits(numBits)

        if numBits != 0 and (value >> (numBits - 1)) != 0:
            # signed
            return value - (1 << numBits)
        else:
            # unsigned
            return value

    def readVarInt16(self) -> int:
        """
        Reads variable 16-bit signed integer value from the bit stream.

        :returns: Variable 16-bit signed integer value.
        """

        byte = self.readBits(8) # byte 1
        sign = byte & VARINT_SIGN_1
        result = byte & VARINT_BYTE_1
        if byte & VARINT_HAS_NEXT_1 == 0:
            return -result if sign != 0 else result

        result = (result << 8) | self.readBits(8) # byte 2
        return -result if sign else result

    def readVarInt32(self) -> int:
        """
        Reads variable 32-bit signed integer value from the bit stream.

        :returns: Variable 32-bit signed integer value.
        """

        byte = self.readBits(8) # byte 1
        sign = byte & VARINT_SIGN_1
        result = byte & VARINT_BYTE_1
        if byte & VARINT_HAS_NEXT_1 == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 2
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 3
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        result = result << 8 | self.readBits(8) # byte 4
        return -result if sign else result

    def readVarInt64(self) -> int:
        """
        Reads variable 64-bit signed integer value from the bit stream.

        :returns: Variable 64-bit signed integer value.
        """

        byte = self.readBits(8) # byte 1
        sign = byte & VARINT_SIGN_1
        result = byte & VARINT_BYTE_1
        if byte & VARINT_HAS_NEXT_1 == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 2
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 3
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 4
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 5
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 6
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 7
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        result = result << 8 | self.readBits(8) # byte 8
        return -result if sign else result

    def readVarInt(self) -> int:
        """
        Reads variable signed integer value (up to 9 bytes) from the bit stream.

        :returns: Variable signed integer value (up to 9 bytes).
        """

        byte = self.readBits(8) # byte 1
        sign = byte & VARINT_SIGN_1
        result = byte & VARINT_BYTE_1
        if byte & VARINT_HAS_NEXT_1 == 0:
            return (INT64_MIN if result == 0 else -result) if sign else result

        byte = self.readBits(8) # byte 2
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 3
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 4
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 5
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 6
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 7
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        byte = self.readBits(8) # byte 8
        result = result << 7 | (byte & VARINT_BYTE_N)
        if byte & VARINT_HAS_NEXT_N == 0:
            return -result if sign else result

        result = result << 8 | self.readBits(8) # byte 9
        return -result if sign else result

    def readVarUInt16(self) -> int:
        """
        Reads variable 16-bit unsigned integer value from the bit stream.

        :returns: Variable 16-bit unsigned integer value.
        """

        byte = self.readBits(8) # byte 1
        result = byte & VARUINT_BYTE
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        result = result << 8 | self.readBits(8) # byte 2
        return result

    def readVarUInt32(self) -> int:
        """
        Reads variable 32-bit unsigned integer value from the bit stream.

        :returns: Variable 32-bit unsigned integer value.
        """

        byte = self.readBits(8) # byte 1
        result = byte & VARUINT_BYTE
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 2
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 3
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        result = result << 8 | self.readBits(8) # byte 4
        return result

    def readVarUInt64(self) -> int:
        """
        Reads variable 64-bit unsigned integer value from the bit stream.

        :returns: Variable 64-bit unsigned integer value.
        """

        byte = self.readBits(8) # byte 1
        result = byte & VARUINT_BYTE
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 2
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 3
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 4
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 5
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 6
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 7
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        result = result << 8 | self.readBits(8) # byte 8
        return result

    def readVarUInt(self) -> int:
        """
        Reads variable unsigned integer value (up to 9 bytes) from the bit stream.

        :returns: Variable unsigned integer value (up to 9 bytes).
        """

        byte = self.readBits(8) # byte 1
        result = byte & VARUINT_BYTE
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 2
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 3
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 4
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 5
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 6
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 7
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 8
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        result = result << 8 | self.readBits(8) # byte 9
        return result

    def readVarSize(self) -> int:
        """
        Reads variable size integer value from the bit stream.

        :returns: Variable size integer value.
        :raises PythonRuntimeException: If read variable size integer is out of range.
        """

        byte = self.readBits(8) # byte 1
        result = byte & VARUINT_BYTE
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 2
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 3
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        byte = self.readBits(8) # byte 4
        result = result << 7 | (byte & VARUINT_BYTE)
        if byte & VARUINT_HAS_NEXT == 0:
            return result

        result = result << 8 | self.readBits(8) # byte 5
        if result > VARSIZE_MAX_VALUE:
            raise PythonRuntimeException("BitStreamReader: Read value '%d' is out of range for varsize type!" %
                                         result)

        return result

    def readFloat16(self) -> float:
        """
        Read 16-bits from the stream as a float value encoded according to IEEE 754 binary16.

        :returns: Read float value.
        :raises PythonRuntimeException: If the reading goes behind the stream.
        """

        return convertUInt16ToFloat(self.readBits(16))

    def readFloat32(self) -> float:
        """
        Read 32-bits from the stream as a float value encoded according to IEEE 754 binary32.

        :returns: Read float value.
        :raises PythonRuntimeException: If the reading goes behind the stream.
        """

        return convertUInt32ToFloat(self.readBits(32))

    def readFloat64(self) -> float:
        """
        Read 64-bits from the stream as a float value encoded according to IEEE 754 binary64.

        :returns: Read float value.
        :raises PythonRuntimeException: If the reading goes behind the stream.
        """

        return convertUInt64ToFloat(self.readBits(64))

    def readString(self) -> str:
        """
        Reads string from the stream.

        :returns: Read string.
        :raises PythonRuntimeException: If the reading goes behind the stream.
        """

        length = self.readVarSize()
        value = bytearray()
        for _ in range(length):
            value.append(self.readBits(8))

        return value.decode("utf-8")

    def readBool(self) -> bool:
        """
        Reads single bit as a bool value.

        :returns: Read bool values.
        :raises PythonRuntimeException: If the reading goes behind the stream.
        """

        return self.readBits(1) != 0

    def readBitBuffer(self) -> BitBuffer:
        """
        Reads a bit buffer from the stream.

        :returns: Read bit buffer.
        :raises PythonRuntimeException: If the reading goes behind the stream.
        """

        bitSize = self.readVarSize()
        numBytesToRead = bitSize // 8
        numRestBits = bitSize - numBytesToRead * 8
        byteSize = (bitSize + 7) // 8
        readBuffer = bytearray(byteSize)
        beginBitPosition = self._bitPosition
        if (beginBitPosition & 0x07) != 0:
            # we are not aligned to byte
            for i in range(numBytesToRead):
                readBuffer[i] = self.readBits(8)
        else:
            # we are aligned to byte
            self.setBitPosition(beginBitPosition + numBytesToRead * 8)
            beginBytePosition = beginBitPosition // 8
            readBuffer[0:numBytesToRead] = self._buffer[beginBytePosition:beginBytePosition + numBytesToRead]

        if numRestBits != 0:
            readBuffer[numBytesToRead] = self.readBits(numRestBits) << (8 - numRestBits)

        return BitBuffer(readBuffer, bitSize)

    def getBitPosition(self) -> int:
        """
        Gets current bit position.

        :returns: Current bit position.
        """

        return self._bitPosition

    def setBitPosition(self, bitPosition: int) -> None:
        """
        Sets bit position.

        :param bitPosition: New bit position.
        :raises PythonRuntimeException: If the position is not within the stream.
        """

        if bitPosition < 0:
            raise PythonRuntimeException("BitStreamReader: Cannot set negative bit position!")
        if bitPosition > len(self._buffer) * 8:
            raise PythonRuntimeException("BitStreamReader: Setting bit position behind the stream!")

        self._bitPosition = bitPosition

    def alignTo(self, alignment: int) -> None:
        """
        Aligns the bit position according to the aligning value.

        :param alignment: An aligning value to use.
        :raises PythonRuntimeException: If the aligning moves behind the stream."
        """

        offset = self._bitPosition % alignment
        if offset != 0:
            self.setBitPosition(self._bitPosition + alignment - offset)

    def getBufferBitSize(self) -> int:
        """
        Gets size of the underlying buffer in bits.

        :returns: Buffer bit size.
        """

        return self._bitSize

VARINT_SIGN_1 = 0x80
VARINT_BYTE_1 = 0x3f
VARINT_BYTE_N = 0x7f
VARINT_HAS_NEXT_1 = 0x40
VARINT_HAS_NEXT_N = 0x80
VARUINT_BYTE = 0x7f
VARUINT_HAS_NEXT = 0x80
VARSIZE_MAX_VALUE = (1 << 31) - 1



"""
The module provides help methods for manipulation with float numbers.

The following float formats defined by IEEE 754 standard are supported:

* half precision float point format (https://en.wikipedia.org/wiki/Half-precision_floating-point_format)
* single precision float point format (https://en.wikipedia.org/wiki/Single-precision_floating-point_format)
* double precision float point format (https://en.wikipedia.org/wiki/Double-precision_floating-point_format)
"""

import struct

def convertUInt16ToFloat(float16Value: int) -> float:
    """
    Converts 16-bit float stored as an integer value to python native float.

    :param float16Value: Half precision float value stored as an integer value to convert.
    :returns: Converted python native float.
    """

    # decompose half precision float (float16)
    sign16Shifted = (float16Value & FLOAT16_SIGN_MASK)
    exponent16 = (float16Value & FLOAT16_EXPONENT_MASK) >> FLOAT16_EXPONENT_BIT_POSITION
    significand16 = (float16Value & FLOAT16_SIGNIFICAND_MASK)

    # calculate significand for single precision float (float32)
    significand32 = significand16 << (FLOAT32_SIGNIFICAND_NUM_BITS - FLOAT16_SIGNIFICAND_NUM_BITS)

    # calculate exponent for single precision float (float32)
    if exponent16 == 0:
        if significand32 != 0:
            # subnormal (denormal) number will be normalized
            exponent32 = (1 + FLOAT32_EXPONENT_BIAS -
                          FLOAT16_EXPONENT_BIAS) # exp is initialized by -14
            # shift significand until leading bit overflows into exponent bit
            while (significand32 & (FLOAT32_SIGNIFICAND_MASK + 1)) == 0:
                exponent32 = exponent32 - 1
                significand32 <<= 1

            # mask out overflowed leading bit from significand (normalized has implicit leading bit 1)
            significand32 &= FLOAT32_SIGNIFICAND_MASK
        else:
            # zero
            exponent32 = 0
    elif exponent16 == FLOAT16_EXPONENT_INFINITY_NAN:
        # infinity or NaN
        exponent32 = FLOAT32_EXPONENT_INFINITY_NAN
    else:
        # normal number
        exponent32 = exponent16 - FLOAT16_EXPONENT_BIAS + FLOAT32_EXPONENT_BIAS

    # compose single precision float (float32)
    sign32Shifted = sign16Shifted << (FLOAT32_SIGN_BIT_POSITION - FLOAT16_SIGN_BIT_POSITION)
    exponent32Shifted = exponent32 << FLOAT32_EXPONENT_BIT_POSITION
    float32Value = sign32Shifted | exponent32Shifted | significand32

    # convert it to float
    return convertUInt32ToFloat(float32Value)

def convertFloatToUInt16(float64: float) -> int:
    """
    Converts python native float to 16-bit float stored as integer value.

    :param float64: Python native float to convert.
    :returns: Converted half precision float value stored as an integer value.
    """

    float32Value = convertFloatToUInt32(float64)

    # decompose single precision float (float32)
    sign32Shifted = (float32Value & FLOAT32_SIGN_MASK)
    exponent32 = (float32Value & FLOAT32_EXPONENT_MASK) >> FLOAT32_EXPONENT_BIT_POSITION
    significand32 = (float32Value & FLOAT32_SIGNIFICAND_MASK)

    # calculate significand for half precision float (float16)
    significand16 = significand32 >> (FLOAT32_SIGNIFICAND_NUM_BITS - FLOAT16_SIGNIFICAND_NUM_BITS)

    # calculate exponent for half precision float (float16)
    needsRounding = False
    if exponent32 == 0:
        if significand32 != 0:
            # subnormal (denormal) number will be zero
            significand16 = 0
        exponent16 = 0
    elif exponent32 == FLOAT32_EXPONENT_INFINITY_NAN:
        # infinity or NaN
        exponent16 = FLOAT16_EXPONENT_INFINITY_NAN
    else:
        # normal number
        signedExponent16 = exponent32 - FLOAT32_EXPONENT_BIAS + FLOAT16_EXPONENT_BIAS
        if signedExponent16 > FLOAT16_EXPONENT_INFINITY_NAN:
            # exponent overflow, set infinity or NaN
            exponent16 = FLOAT16_EXPONENT_INFINITY_NAN
        elif signedExponent16 <= 0:
            # exponent underflow
            if signedExponent16 <= -FLOAT16_SIGNIFICAND_NUM_BITS:
                # too big underflow, set to zero
                exponent16 = 0
                significand16 = 0
            else:
                # we can still use subnormal numbers
                exponent16 = 0
                fullSignificand32 = significand32 | (FLOAT32_SIGNIFICAND_MASK + 1)
                significandShift = 1 - signedExponent16
                significand16 = fullSignificand32 >> (FLOAT32_SIGNIFICAND_NUM_BITS -
                                                      FLOAT16_SIGNIFICAND_NUM_BITS + significandShift)

                needsRounding = ((fullSignificand32 >>
                                  (FLOAT32_SIGNIFICAND_NUM_BITS - FLOAT16_SIGNIFICAND_NUM_BITS +
                                   significandShift - 1)) & 1) != 0
        else:
            # exponent ok
            exponent16 = signedExponent16
            needsRounding = ((significand32 >> (FLOAT32_SIGNIFICAND_NUM_BITS -
                                                FLOAT16_SIGNIFICAND_NUM_BITS - 1)) & 1) != 0

    # compose half precision float (float16)
    sign16Shifted = sign32Shifted >> (FLOAT32_SIGN_BIT_POSITION - FLOAT16_SIGN_BIT_POSITION)
    exponent16Shifted = exponent16 << FLOAT16_EXPONENT_BIT_POSITION
    float16Value = sign16Shifted | exponent16Shifted | significand16

    # check rounding
    if needsRounding:
        float16Value += 1   # might overflow to infinity

    return float16Value

def convertUInt32ToFloat(float32Value: int) -> float:
    """
    Converts 32-bit float stored as an integer value to python native float.

    :param float32Value: Single precision float value stored as an integer value to convert.
    :returns: Converted python native float.
    """

    float32ValueInBytes = float32Value.to_bytes(4, byteorder="big")

    return struct.unpack('>f', float32ValueInBytes)[0]

def convertFloatToUInt32(float64: float) -> int:
    """
    Converts python native float to 32-bit float stored as integer value.

    :param float64: Python native float to convert.
    :returns: Converted single precision float value stored as an integer value.
    """

    float32ValueInBytes = struct.pack('>f', float64)

    return int.from_bytes(float32ValueInBytes, byteorder="big")

def convertUInt64ToFloat(float64Value: int) -> float:
    """
    Converts 64-bit float stored as an integer value to python native float.

    :param float64Value: Double precision float value stored as an integer value to convert.
    :returns: Converted python native float.
    """

    float64ValueInBytes = float64Value.to_bytes(8, byteorder="big")

    return struct.unpack('>d', float64ValueInBytes)[0]

def convertFloatToUInt64(float64: float) -> int:
    """
    Converts python native float to 64-bit float stored as integer value.

    :param float64: Python native float to convert.
    :returns: Converted double precision float value stored as an integer value.
    """

    float64ValueInBytes = struct.pack('>d', float64)

    return int.from_bytes(float64ValueInBytes, byteorder="big")

FLOAT16_SIGN_MASK = 0x8000
FLOAT16_EXPONENT_MASK = 0x7C00
FLOAT16_SIGNIFICAND_MASK = 0x03FF

FLOAT16_SIGN_BIT_POSITION = 15
FLOAT16_EXPONENT_BIT_POSITION = 10

FLOAT16_SIGNIFICAND_NUM_BITS = FLOAT16_EXPONENT_BIT_POSITION

FLOAT16_EXPONENT_INFINITY_NAN = 0x001F
FLOAT16_EXPONENT_BIAS = 15

FLOAT32_SIGN_MASK = 0x80000000
FLOAT32_EXPONENT_MASK = 0x7F800000
FLOAT32_SIGNIFICAND_MASK = 0x007FFFFF

FLOAT32_SIGN_BIT_POSITION = 31
FLOAT32_EXPONENT_BIT_POSITION = 23

FLOAT32_SIGNIFICAND_NUM_BITS = FLOAT32_EXPONENT_BIT_POSITION

FLOAT32_EXPONENT_INFINITY_NAN = 0x00FF
FLOAT32_EXPONENT_BIAS = 127



"""
The module provides implementation of zserio built-in operators.
"""

def getNumBits(numValues: int) -> int:
    """
    Gets the minimum number of bits required to encode given number of different values.

    This method implements zserio built-in operator numBits.

    :param numValues: The number of different values from which to calculate number of bits.
    :returns: Number of bits required to encode numValues different values.
    """

    if numValues == 0:
        return 0
    if numValues == 1:
        return 1

    return (numValues - 1).bit_length()



"""
The module implements abstraction for holding bit sequence.
"""

import typing

from zserio.exception import PythonRuntimeException
from zserio.hashcode import HASH_SEED
from zserio.hashcode import calcHashCode

class BitBuffer:
    """
    Bit buffer.

    Because bit buffer size does not have to be byte aligned (divisible by 8), it's possible that not all bits
    of the last byte are used. In this case, only most significant bits of the corresponded size are used.
    """

    def __init__(self, buffer: bytes, bitSize: typing.Optional[int] = None) -> None:
        """
        Constructs bit buffer from bytes buffer and bit size.

        :param buffer: Bytes-like buffer to construct from.
        :param bitSize: Number of bits stored in buffer to use.
        :raises PythonRuntimeException: If the numBits is invalid number of the reading goes behind the stream.
        """

        if bitSize is None:
            bitSize = len(buffer) * 8
        elif len(buffer) * 8 < bitSize:
            raise PythonRuntimeException("BitBuffer: Bit size %d out of range for given buffer byte "
                                         "size %d!" % (bitSize, len(buffer)))
        self._buffer = buffer # type: bytes
        self._bitSize = bitSize # type: int

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BitBuffer):
            return False

        if self._bitSize != other._bitSize:
            return False

        byteSize = self.getByteSize()
        if byteSize > 0:
            if byteSize > 1:
                if self._buffer[0:byteSize - 1] != other._buffer[0:byteSize - 1]:
                    return False

            if self._getMaskedLastByte() != other._getMaskedLastByte():
                return False

        return True

    def __hash__(self) -> int:
        result = HASH_SEED
        byteSize = self.getByteSize()
        if byteSize > 0:
            if byteSize > 1:
                for element in self._buffer[0:byteSize - 1]:
                    result = calcHashCode(result, hash(element))

            result = calcHashCode(result, hash(self._getMaskedLastByte()))

        return result

    def getBuffer(self) -> bytes:
        """
        Gets the underlying byte buffer.

        Not all bits of the last byte must be used.

        :returns: The underlying byte buffer.
        """
        return self._buffer

    def getBitSize(self) -> int:
        """
        Gets the number of bits stored in the bit buffer.

        :returns: Size of the bit buffer in bits.
        """
        return self._bitSize

    def getByteSize(self) -> int:
        """
        Gets the number of bytes stored in the bit buffer.

        :returns: Size of the bit buffer in bytes.
        """
        return (self._bitSize + 7) // 8

    def _getMaskedLastByte(self) -> int:
        roundedByteSize = self._bitSize // 8
        lastByteBits = self._bitSize - 8 * roundedByteSize

        return (self._buffer[roundedByteSize - 1] if lastByteBits == 0 else
                self._buffer[roundedByteSize] & (0xFF << (8 - lastByteBits)))



"""
The module provides utility methods for hash code calculation.
"""

HASH_SEED = 23

def calcHashCode(seedValue: int, value: int) -> int:
    """
    Calculates hash code of the value using seed value.

    :param seedValue: Seed value (current hash code).
    :param value: Value for which to calculate hash code.
    :returns: Calculated hash code.
    """

    return (HASH_PRIME_NUMBER * seedValue + value) & 0xFFFFFFFF

HASH_PRIME_NUMBER = 37



"""
The module implements abstraction for writing data to the bit stream.
"""

from zserio.bitbuffer import BitBuffer
from zserio.bitsizeof import (getBitSizeOfVarInt16, getBitSizeOfVarInt32,
                              getBitSizeOfVarInt64, getBitSizeOfVarInt,
                              getBitSizeOfVarUInt16, getBitSizeOfVarUInt32,
                              getBitSizeOfVarUInt64, getBitSizeOfVarUInt,
                              getBitSizeOfVarSize, INT64_MIN)
from zserio.exception import PythonRuntimeException
from zserio.float import convertFloatToUInt16, convertFloatToUInt32, convertFloatToUInt64

class BitStreamWriter:
    """
    Bit stream writer using bytearray.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """

        self._byteArray = bytearray() # type: bytearray
        self._bitPosition = 0 # type: int

    def writeBits(self, value: int, numBits: int) -> None:
        """
        Writes the given value with the given number of bits to the underlying storage.

        :param value: Value to write.
        :param numBits: Number of bits to write.
        :raises PythonRuntimeException: If the value is out of the range or if the number of bits is invalid.
        """

        if numBits <= 0:
            raise PythonRuntimeException("BitStreamWriter: numBits '%d' is less than 1!" % numBits)

        minValue = 0
        maxValue = (1 << numBits) - 1
        if value < minValue or value > maxValue:
            raise PythonRuntimeException("BitStreamWriter: Value '%d' is out of the range <%d,%d>!" %
                                         (value, minValue, maxValue))

        self._writeBitsImpl(value, numBits, signed=False)

    def writeSignedBits(self, value: int, numBits: int) -> None:
        """
        Writes the given signed value with the given number of bits to the underlying storage.
        Provided for convenience.

        :param value: Signed value to write.
        :param numBits: Number of bits to write.
        :raises PythonRuntimeException: If the value is out of the range or if the number of bits is invalid.
        """

        if numBits <= 0:
            raise PythonRuntimeException("BitStreamWriter: numBits '%d' is less than 1!" % numBits)

        minValue = -(1 << (numBits - 1))
        maxValue = (1 << (numBits - 1)) - 1
        if value < minValue or value > maxValue:
            raise PythonRuntimeException("BitStreamWriter: Value '%d' is out of the range <%d,%d>!" %
                                         (value, minValue, maxValue))

        self._writeBitsImpl(value, numBits, signed=True)

    def writeVarInt16(self, value: int) -> None:
        """
        Writes a variable 16-bit signed integer value to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        self._writeVarNum(value, 2, getBitSizeOfVarInt16(value) // 8, isSigned=True)

    def writeVarInt32(self, value: int) -> None:
        """
        Writes a variable 32-bit signed integer value to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        self._writeVarNum(value, 4, getBitSizeOfVarInt32(value) // 8, isSigned=True)

    def writeVarInt64(self, value: int) -> None:
        """
        Writes a variable 16-bit signed integer value to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        self._writeVarNum(value, 8, getBitSizeOfVarInt64(value) // 8, isSigned=True)

    def writeVarInt(self, value: int) -> None:
        """
        Writes a variable signed integer value (up to 9 bytes) to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        if value == INT64_MIN:
            self.writeBits(0x80, 8) # INT64_MIN is stored as -0
        else:
            self._writeVarNum(value, 9, getBitSizeOfVarInt(value) // 8, isSigned=True)

    def writeVarUInt16(self, value: int) -> None:
        """
        Writes a variable 16-bit unsigned integer value to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        self._writeVarNum(value, 2, getBitSizeOfVarUInt16(value) // 8, isSigned=False)

    def writeVarUInt32(self, value: int) -> None:
        """
        Writes a variable 32-bit unsigned integer value to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        self._writeVarNum(value, 4, getBitSizeOfVarUInt32(value) // 8, isSigned=False)

    def writeVarUInt64(self, value: int) -> None:
        """
        Writes a variable 16-bit unsigned integer value to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        self._writeVarNum(value, 8, getBitSizeOfVarUInt64(value) // 8, isSigned=False)

    def writeVarUInt(self, value: int) -> None:
        """
        Writes a variable unsigned integer value (up to 9 bytes) to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        self._writeVarNum(value, 9, getBitSizeOfVarUInt(value) // 8, isSigned=False)

    def writeVarSize(self, value: int) -> None:
        """
        Writes a variable size integer value to the underlying storage.

        :param value: Value to write.
        :raises PythonRuntimeException: If the value is out of the range.
        """

        self._writeVarNum(value, 5, getBitSizeOfVarSize(value) // 8, isSigned=False)

    def writeFloat16(self, value: float) -> None:
        """
        Writes a 16-bit float value to the underlying storage according to IEEE 754 binary16.

        :param value: Float value to write.
        """

        self.writeBits(convertFloatToUInt16(value), 16)

    def writeFloat32(self, value: float) -> None:
        """
        Writes a 32-bit float value to the underlying storage according to IEEE 754 binary32.

        :param value: Float value to write.
        """

        self.writeBits(convertFloatToUInt32(value), 32)

    def writeFloat64(self, value: float) -> None:
        """
        Writes a 64-bit float value to the underlying storage according to IEEE 754 binary64.

        :param value: Float value to write.
        """

        self.writeBits(convertFloatToUInt64(value), 64)

    def writeString(self, string: str) -> None:
        """
        Writes the given string to the underlying storage in UTF-8 encoding. Length of the string is written
        as varuint64 at the beginning.

        :param string: String to write.
        """

        stringBytes = string.encode("utf-8")
        self.writeVarSize(len(stringBytes))
        for stringByte in stringBytes:
            self.writeBits(stringByte, 8)

    def writeBool(self, value: bool) -> None:
        """
        Writes bool in a single bit.

        :param value: Bool value to write.
        """

        self.writeBits(1 if value else 0, 1)

    def writeBitBuffer(self, bitBuffer: BitBuffer) -> None:
        """
        Writes a bit buffer to the underlying storage. Length of the bit buffer is written as varuint64
        at the beginning.

        :param bitBuffer: Bit buffer to write.
        """

        bitSize = bitBuffer.getBitSize()
        self.writeVarSize(bitSize)

        writeBuffer = bitBuffer.getBuffer()
        numBytesToWrite = bitSize // 8
        numRestBits = bitSize - numBytesToWrite * 8
        beginBitPosition = self._bitPosition
        if (beginBitPosition & 0x07) != 0:
            # we are not aligned to byte
            for i in range(numBytesToWrite):
                self.writeBits(writeBuffer[i], 8)
        else:
            # we are aligned to byte
            self._byteArray += writeBuffer[0:numBytesToWrite]
            self._bitPosition += numBytesToWrite * 8

        if numRestBits > 0:
            self.writeBits(writeBuffer[numBytesToWrite] >> (8 - numRestBits), numRestBits)

    def getByteArray(self) -> bytes:
        """
        Gets internal bytearray.

        :returns: Underlying bytearray object.
        """

        return self._byteArray

    def toFile(self, filename: str) -> None:
        """
        Writes underlying bytearray to binary file.

        :param filename: File to write.
        """

        with open(filename, "wb") as file:
            file.write(self._byteArray)

    def getBitPosition(self) -> int:
        """
        Gets current bit position.

        :returns: Current bit position.
        """

        return self._bitPosition

    def alignTo(self, alignment: int) -> None:
        """
        Aligns the bit position according to the aligning value.

        :param alignment: An aligning value to use.
        """

        offset = self._bitPosition % alignment
        if offset != 0:
            self.writeBits(0, alignment - offset)

    def _writeBitsImpl(self, value: int, numBits: int, *, signed: bool) -> None:
        bufferLastByteBits = self._bitPosition % 8
        bufferFreeBits = (8 - bufferLastByteBits) if bufferLastByteBits != 0 else 0
        valueFirstByteBits = numBits % 8 or 8
        if valueFirstByteBits <= bufferFreeBits:
            leftShift = bufferFreeBits - valueFirstByteBits
        else:
            leftShift = bufferFreeBits + 8 - valueFirstByteBits
        value <<= leftShift
        numBytes = (numBits + leftShift + 7) // 8
        valueBytes = value.to_bytes(numBytes, byteorder='big', signed=signed)
        if bufferFreeBits == 0:
            self._byteArray.extend(valueBytes)
        else:
            valueFirstByte = valueBytes[0] & ((1 << bufferFreeBits) - 1)
            self._byteArray[-1] |= valueFirstByte
            self._byteArray.extend(valueBytes[1:])

        self._bitPosition += numBits

    def _writeVarNum(self, value: int, maxVarBytes: int, numVarBytes: int, *, isSigned: bool) -> None:
        absValue = abs(value)
        hasMaxByteRange = (numVarBytes == maxVarBytes)
        for i in range(numVarBytes):
            byte = 0x00
            numBits = 8
            hasNextByte = (i < numVarBytes - 1)
            hasSignBit = (isSigned and i == 0)
            if hasSignBit:
                if value < 0:
                    byte |= 0x80
                numBits -= 1
            if hasNextByte:
                numBits -= 1
                byte |= (1 << numBits) # use bit 6 if signed bit is present, use bit 7 otherwise
            else: # this is the last byte
                if not hasMaxByteRange: # next byte indicator is not used in last byte in case of max byte range
                    numBits -= 1

            shiftBits = (numVarBytes - (i + 1)) * 7 + (1 if hasMaxByteRange and hasNextByte else 0)
            byte |= (absValue >> shiftBits) & VAR_NUM_BIT_MASKS[numBits - 1]
            self.writeBits(byte, 8)

VAR_NUM_BIT_MASKS = [0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff]



"""
The module provides classes for Zserio services.
"""

import typing

from zserio.exception import PythonRuntimeException

class ServiceInterface:
    """
    Generic interface for all Zserio services.
    """

    def callMethod(self, methodName: str, requestData: bytes, context: typing.Any = None) -> bytes:
        """
        Calls method with the given name synchronously.

        :param methodName: Name of the service method to call.
        :param requestData: Request data to be passed to the method.
        :param context: Context specific for particular service.
        :returns: Response data.
        :raises ServiceException: If the call fails.
        """
        raise NotImplementedError()

class ServiceException(PythonRuntimeException):
    """
    Exception thrown in case of an error in Zserio Service
    """



"""
The module implements abstraction for arrays used by Zserio python extension.
"""

import typing

from zserio.bitposition import alignTo
from zserio.bitsizeof import (getBitSizeOfVarUInt16, getBitSizeOfVarUInt32,
                              getBitSizeOfVarUInt64, getBitSizeOfVarUInt,
                              getBitSizeOfVarSize, getBitSizeOfVarInt16,
                              getBitSizeOfVarInt32, getBitSizeOfVarInt64,
                              getBitSizeOfVarInt, getBitSizeOfString,
                              getBitSizeOfBitBuffer)
from zserio.bitreader import BitStreamReader
from zserio.bitwriter import BitStreamWriter
from zserio.bitbuffer import BitBuffer
from zserio.hashcode import calcHashCode, HASH_SEED
from zserio.exception import PythonRuntimeException

class Array:
    """
    Abstraction for arrays to which Zserio arrays are mapped in python.
    """

    def __init__(self,
                 arrayTraits: typing.Any,
                 rawArray: typing.Optional[typing.List] = None,
                 *,
                 isAuto: bool = False,
                 isImplicit: bool = False,
                 setOffsetMethod: typing.Optional[typing.Callable[[int, int], None]] = None,
                 checkOffsetMethod: typing.Optional[typing.Callable[[int, int], None]] = None) -> None:
        """
        Constructor.

        :param arrayTraits: Array traits which specify the array type.
        :param rawArray: Native python list which will be hold by this abstraction.
        :param isAuto: True if mapped Zserio array is auto array.
        :param isImplicit: True if mapped Zserio array is implicit array.
        :param setOffsetMethod:  Set offset method if mapped Zserio array is indexed offset array.
        :param checkOffsetMethod: Check offset method if mapped Zserio array is indexed offset array.
        """

        self._rawArray = [] if rawArray is None else rawArray # type: typing.List
        self._arrayTraits = arrayTraits # type: typing.Any
        self._isAuto = isAuto # type: bool
        self._isImplicit = isImplicit # type: bool
        self._setOffsetMethod = setOffsetMethod # typing.Optional[typing.Callable[[int, int], None]]
        self._checkOffsetMethod = checkOffsetMethod # typing.Optional[typing.Callable[[int, int], None]]

    @classmethod
    def fromReader(cls: typing.Type['Array'],
                   arrayTraits: typing.Any,
                   reader: BitStreamReader,
                   size: int = 0,
                   *,
                   isAuto: bool = False,
                   isImplicit: bool = False,
                   setOffsetMethod: typing.Optional[typing.Callable[[int, int], None]] = None,
                   checkOffsetMethod: typing.Optional[typing.Callable[[int, int], None]] = None) -> 'Array':
        """
        Constructs array and reads elements from the given bit stream reader.

        :param arrayTraits: Array traits which specify the array type.
        :param reader: Bit stream from which to read.
        :param size: Number of elements to read or None in case of implicit or auto arrays.
        :param rawArray: Native python list which will be hold by this abstraction.
        :param isAuto: True if mapped Zserio array is auto array.
        :param isImplicit: True if mapped Zserio array is implicit array.
        :param setOffsetMethod:  Set offset method if mapped Zserio array is indexed offset array.
        :param checkOffsetMethod: Check offset method if mapped Zserio array is indexed offset array.
        :returns: Array instance filled using given bit stream reader.
        """

        instance = cls(arrayTraits, isAuto=isAuto, isImplicit=isImplicit, setOffsetMethod=setOffsetMethod,
                       checkOffsetMethod=checkOffsetMethod)
        instance.read(reader, size)

        return instance

    def __eq__(self, other: object) -> bool:
        # it's enough to check only rawArray because compound types which call this are always the same type
        if isinstance(other, Array):
            return self._rawArray == other._rawArray

        return False

    def __hash__(self) -> int:
        hashCode = HASH_SEED
        for element in self._rawArray:
            hashCode = calcHashCode(hashCode, hash(element))

        return hashCode

    def __len__(self) -> int:
        return len(self._rawArray)

    def __getitem__(self, key: int) -> typing.Any:
        return self._rawArray[key]

    def __setitem__(self, key: int, value: typing.Any) -> None:
        self._rawArray[key] = value

    def getRawArray(self) -> typing.List:
        """
        Gets raw array.

        :returns: Native python list which is hold by the array.
        """

        return self._rawArray


    def bitSizeOf(self, bitPosition: int) -> int:
        """
        Returns length of array stored in the bit stream in bits.

        :param bitPosition: Current bit stream position.
        :returns: Length of the array stored in the bit stream in bits.
        """

        endBitPosition = bitPosition
        size = len(self._rawArray)
        if self._isAuto:
            endBitPosition += getBitSizeOfVarSize(size)

        if self._arrayTraits.HAS_BITSIZEOF_CONSTANT and size > 0:
            elementSize = self._arrayTraits.bitSizeOf()
            if self._setOffsetMethod is None:
                endBitPosition += size * elementSize
            else:
                endBitPosition += elementSize + (size - 1) * alignTo(8, elementSize)
        else:
            for element in self._rawArray:
                if self._setOffsetMethod is not None:
                    endBitPosition = alignTo(8, endBitPosition)
                endBitPosition += self._arrayTraits.bitSizeOf(endBitPosition, element)

        return endBitPosition - bitPosition

    def initializeOffsets(self, bitPosition: int) -> int:
        """
        Initializes indexed offsets for the array.

        :param bitPosition: Current bit stream position.
        :returns: Updated bit stream position which points to the first bit after the array.
        """

        endBitPosition = bitPosition
        size = len(self._rawArray)
        if self._isAuto:
            endBitPosition += getBitSizeOfVarSize(size)

        for index in range(size):
            if self._setOffsetMethod is not None:
                endBitPosition = alignTo(8, endBitPosition)
                self._setOffsetMethod(index, endBitPosition)
            endBitPosition = self._arrayTraits.initializeOffsets(endBitPosition, self._rawArray[index])

        return endBitPosition

    def read(self, reader: BitStreamReader, size: int = 0) -> None:
        """
        Reads array from the bit stream.

        :param reader: Bit stream from which to read.
        :param size: Number of elements to read or None in case of implicit or auto arrays.

        :raises PythonRuntimeException: If the array does not have elements with constant bit size.
        """

        self._rawArray.clear()

        if self._isImplicit:
            if not self._arrayTraits.HAS_BITSIZEOF_CONSTANT:
                raise PythonRuntimeException("Array: Implicit array elements must have constant bit size!")

            elementSize = self._arrayTraits.bitSizeOf()
            remainingBits = reader.getBufferBitSize() - reader.getBitPosition()
            readSize = remainingBits // elementSize
            for index in range(readSize):
                self._rawArray.append(self._arrayTraits.read(reader, index))
        else:
            if self._isAuto:
                readSize = reader.readVarSize()
            else:
                readSize = size

            for index in range(readSize):
                if self._checkOffsetMethod is not None:
                    reader.alignTo(8)
                    self._checkOffsetMethod(index, reader.getBitPosition())
                self._rawArray.append(self._arrayTraits.read(reader, index))

    def write(self, writer: BitStreamWriter) -> None:
        """
        Writes array to the bit stream.

        :param writer: Bit stream where to write.
        """

        size = len(self._rawArray)
        if self._isAuto:
            writer.writeVarSize(size)

        for index in range(size):
            if self._checkOffsetMethod is not None:
                writer.alignTo(8)
                self._checkOffsetMethod(index, writer.getBitPosition())
            self._arrayTraits.write(writer, self._rawArray[index])

class BitFieldArrayTraits:
    """
    Array traits for unsigned fixed integer Zserio types (uint16, uint32, uint64, bit:5, etc...).
    """

    HAS_BITSIZEOF_CONSTANT = True

    def __init__(self, numBits: int) -> None:
        """
        Constructor.

        :param numBits: Number of bits for unsigned fixed integer Zserio type.
        """

        self._numBits = numBits # type: int

    def bitSizeOf(self) -> int:
        """
        Returns length of unsigned fixed integer Zserio type stored in the bit stream in bits.

        :returns: Length of unsigned fixed integer Zserio type in bits.
        """

        return self._numBits

    def initializeOffsets(self, bitPosition: int, _value: int) -> int:
        """
        Initializes indexed offsets for unsigned fixed integer Zserio type.

        :param bitPosition: Current bit stream position.
        :param _value: Not used.
        :returns: Updated bit stream position which points to the first bit after unsigned fixed integer type.
        """

        return bitPosition + self.bitSizeOf()

    def read(self, reader: BitStreamReader, _index: int) -> int:
        """
        Reads unsigned fixed integer Zserio type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readBits(self._numBits)

    def write(self, writer: BitStreamWriter, value: int) -> None:
        """
        Writes unsigned fixed integer Zserio type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Unsigned fixed integer Zserio type to write.
        """

        writer.writeBits(value, self._numBits)

class SignedBitFieldArrayTraits:
    """
    Array traits for signed fixed integer Zserio types (int16, int32, int64, int:5, etc...).
    """

    HAS_BITSIZEOF_CONSTANT = True

    def __init__(self, numBits: int) -> None:
        """
        Constructor.

        :param numBits: Number of bits for signed fixed integer Zserio type.
        """

        self._numBits = numBits

    def bitSizeOf(self) -> int:
        """
        Returns length of signed fixed integer Zserio type stored in the bit stream in bits.

        :returns: Length of signed fixed integer Zserio type in bits.
        """

        return self._numBits

    def initializeOffsets(self, bitPosition: int, _value: int) -> int:
        """
        Initializes indexed offsets for signed fixed integer Zserio type.

        :param bitPosition: Current bit stream position.
        :param _value: Not used.
        :returns: Updated bit stream position which points to the first bit after signed fixed integer type.
        """

        return bitPosition + self.bitSizeOf()

    def read(self, reader: BitStreamReader, _index: int) -> int:
        """
        Reads signed fixed integer Zserio type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readSignedBits(self._numBits)

    def write(self, writer: BitStreamWriter, value: int) -> None:
        """
        Writes signed fixed integer Zserio type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Signed fixed integer Zserio type to write.
        """

        writer.writeSignedBits(value, self._numBits)

class VarUInt16ArrayTraits:
    """
    Array traits for Zserio varuint16 type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varuint16 type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varuint16 type value.
        :returns: Length of given Zserio varuint16 type in bits.
        """

        return getBitSizeOfVarUInt16(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varuint16 type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varuint16 type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varuint16 type.
        """

        return bitPosition + VarUInt16ArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varuint16 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarUInt16()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varuint16 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varuint16 type to write.
        """

        writer.writeVarUInt16(value)

class VarUInt32ArrayTraits:
    """
    Array traits for Zserio varuint32 type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varuint32 type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varuint32 type value.
        :returns: Length of given Zserio varuint32 type in bits.
        """

        return getBitSizeOfVarUInt32(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varuint32 type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varuint32 type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varuint32 type.
        """

        return bitPosition + VarUInt32ArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varuint32 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarUInt32()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varuint32 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varuint32 type to write.
        """

        writer.writeVarUInt32(value)

class VarUInt64ArrayTraits:
    """
    Array traits for Zserio varuint64 type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varuint64 type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varuint64 type value.
        :returns: Length of given Zserio varuint64 type in bits.
        """

        return getBitSizeOfVarUInt64(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varuint64 type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varuint64 type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varuint64 type.
        """

        return bitPosition + VarUInt64ArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varuint64 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarUInt64()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varuint64 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varuint64 type to write.
        """

        writer.writeVarUInt64(value)

class VarUIntArrayTraits:
    """
    Array traits for Zserio varuint type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varuint type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varuint type value.
        :returns: Length of given Zserio varuint type in bits.
        """

        return getBitSizeOfVarUInt(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varuint type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varuint type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varuint type.
        """

        return bitPosition + VarUIntArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varuint type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarUInt()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varuint type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varuint type to write.
        """

        writer.writeVarUInt(value)

class VarSizeArrayTraits:
    """
    Array traits for Zserio varsize type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varsize type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varsize type value.
        :returns: Length of given Zserio varsize type in bits.
        """

        return getBitSizeOfVarSize(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varsize type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varsize type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varsize type.
        """

        return bitPosition + VarSizeArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varsize type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarSize()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varsize type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varsize type to write.
        """

        writer.writeVarSize(value)

class VarInt16ArrayTraits:
    """
    Array traits for Zserio varint16 type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varint16 type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varint16 type value.
        :returns: Length of given Zserio varint16 type in bits.
        """

        return getBitSizeOfVarInt16(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varint16 type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varint16 type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varint16 type.
        """

        return bitPosition + VarInt16ArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varint16 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarInt16()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varint16 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varint16 type to write.
        """

        writer.writeVarInt16(value)

class VarInt32ArrayTraits:
    """
    Array traits for Zserio varint32 type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varint32 type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varint32 type value.
        :returns: Length of given Zserio varint32 type in bits.
        """

        return getBitSizeOfVarInt32(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varint32 type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varint32 type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varint32 type.
        """

        return bitPosition + VarInt32ArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varint32 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarInt32()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varint32 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varint32 type to write.
        """

        writer.writeVarInt32(value)

class VarInt64ArrayTraits:
    """
    Array traits for Zserio varint64 type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varint64 type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varint64 type value.
        :returns: Length of given Zserio varint64 type in bits.
        """

        return getBitSizeOfVarInt64(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varint64 type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varint64 type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varint64 type.
        """

        return bitPosition + VarInt64ArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varint64 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarInt64()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varint64 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varint64 type to write.
        """

        writer.writeVarInt64(value)

class VarIntArrayTraits:
    """
    Array traits for Zserio varint type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: int) -> int:
        """
        Returns length of Zserio varint type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio varint type value.
        :returns: Length of given Zserio varint type in bits.
        """

        return getBitSizeOfVarInt(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: int) -> int:
        """
        Initializes indexed offsets for Zserio varint type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio varint type value.
        :returns: Updated bit stream position which points to the first bit after Zserio varint type.
        """

        return bitPosition + VarIntArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> int:
        """
        Reads Zserio varint type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readVarInt()

    @staticmethod
    def write(writer: BitStreamWriter, value: int) -> None:
        """
        Writes Zserio varint type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio varint type to write.
        """

        writer.writeVarInt(value)

class Float16ArrayTraits:
    """
    Array traits for Zserio float16 type.
    """

    HAS_BITSIZEOF_CONSTANT = True

    @staticmethod
    def bitSizeOf() -> int:
        """
        Returns length of Zserio float16 type stored in the bit stream in bits.

        :returns: Length of Zserio float16 type in bits.
        """

        return 16

    @staticmethod
    def initializeOffsets(bitPosition: int, _value: float) -> int:
        """
        Initializes indexed offsets for Zserio float16 type.

        :param bitPosition: Current bit stream position.
        :param _value: Not used.
        :returns: Updated bit stream position which points to the first bit after Zserio float16 type.
        """

        return bitPosition + Float16ArrayTraits.bitSizeOf()

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> float:
        """
        Reads Zserio float16 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readFloat16()

    @staticmethod
    def write(writer: BitStreamWriter, value: float) -> None:
        """
        Writes Zserio float16 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio float16 type to write.
        """

        writer.writeFloat16(value)

class Float32ArrayTraits:
    """
    Array traits for Zserio float32 type.
    """

    HAS_BITSIZEOF_CONSTANT = True

    @staticmethod
    def bitSizeOf() -> int:
        """
        Returns length of Zserio float32 type stored in the bit stream in bits.

        :returns: Length of Zserio float32 type in bits.
        """

        return 32

    @staticmethod
    def initializeOffsets(bitPosition: int, _value: float) -> int:
        """
        Initializes indexed offsets for Zserio float32 type.

        :param bitPosition: Current bit stream position.
        :param _value: Not used.
        :returns: Updated bit stream position which points to the first bit after Zserio float32 type.
        """

        return bitPosition + Float32ArrayTraits.bitSizeOf()

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> float:
        """
        Reads Zserio float32 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readFloat32()

    @staticmethod
    def write(writer: BitStreamWriter, value: float) -> None:
        """
        Writes Zserio float32 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio float32 type to write.
        """

        writer.writeFloat32(value)

class Float64ArrayTraits:
    """
    Array traits for Zserio float64 type.
    """

    HAS_BITSIZEOF_CONSTANT = True

    @staticmethod
    def bitSizeOf() -> int:
        """
        Returns length of Zserio float64 type stored in the bit stream in bits.

        :returns: Length of Zserio float64 type in bits.
        """

        return 64

    @staticmethod
    def initializeOffsets(bitPosition: int, _value: float) -> int:
        """
        Initializes indexed offsets for Zserio float64 type.

        :param bitPosition: Current bit stream position.
        :param _value: Not used.
        :returns: Updated bit stream position which points to the first bit after Zserio float64 type.
        """

        return bitPosition + Float64ArrayTraits.bitSizeOf()

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> float:
        """
        Reads Zserio float64 type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readFloat64()

    @staticmethod
    def write(writer: BitStreamWriter, value: float) -> None:
        """
        Writes Zserio float64 type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio float64 type to write.
        """

        writer.writeFloat64(value)

class StringArrayTraits:
    """
    Array traits for Zserio string type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition, value: str) -> int:
        """
        Returns length of Zserio string type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio string type value.
        :returns: Length of given Zserio string type in bits.
        """

        return getBitSizeOfString(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: str) -> int:
        """
        Initializes indexed offsets for Zserio string type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio string type value.
        :returns: Updated bit stream position which points to the first bit after Zserio string type.
        """

        return bitPosition + StringArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> str:
        """
        Reads Zserio string type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readString()

    @staticmethod
    def write(writer: BitStreamWriter, value: str) -> None:
        """
        Writes Zserio string type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio string type to write.
        """

        writer.writeString(value)

class BoolArrayTraits:
    """
    Array traits for Zserio bool type.
    """

    HAS_BITSIZEOF_CONSTANT = True

    @staticmethod
    def bitSizeOf() -> int:
        """
        Returns length of Zserio bool type stored in the bit stream in bits.

        :returns: Length of Zserio bool type in bits.
        """

        return 1

    @staticmethod
    def initializeOffsets(bitPosition: int, _value: bool) -> int:
        """
        Initializes indexed offsets for Zserio bool type.

        :param bitPosition: Current bit stream position.
        :param _value: Not used.
        :returns: Updated bit stream position which points to the first bit after Zserio bool type.
        """

        return bitPosition + BoolArrayTraits.bitSizeOf()

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> bool:
        """
        Reads Zserio bool type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readBool()

    @staticmethod
    def write(writer: BitStreamWriter, value: bool) -> None:
        """
        Writes Zserio bool type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio bool type to write.
        """

        writer.writeBool(value)

class BitBufferArrayTraits:
    """
    Array traits for Zserio extern bit buffer type.
    """

    HAS_BITSIZEOF_CONSTANT = False

    @staticmethod
    def bitSizeOf(_bitPosition: int, value: BitBuffer) -> int:
        """
        Returns length of Zserio extern bit buffer type stored in the bit stream in bits.

        :param _bitPosition: Not used.
        :param value: Zserio extern bit buffer type value.
        :returns: Length of given Zserio string type in bits.
        """

        return getBitSizeOfBitBuffer(value)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: BitBuffer) -> int:
        """
        Initializes indexed offsets for Zserio extern bit buffer type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio extern bit buffer type value.
        :returns: Updated bit stream position which points to the first bit after Zserio extern bit buffer type.
        """

        return bitPosition + BitBufferArrayTraits.bitSizeOf(bitPosition, value)

    @staticmethod
    def read(reader: BitStreamReader, _index: int) -> BitBuffer:
        """
        Reads Zserio extern bit buffer type from the bit stream.

        :param reader: Bit stream from which to read.
        :param _index: Not used.
        """

        return reader.readBitBuffer()

    @staticmethod
    def write(writer: BitStreamWriter, value: BitBuffer) -> None:
        """
        Writes Zserio extern bit buffer type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio extern bit buffer type to write.
        """

        writer.writeBitBuffer(value)

class ObjectArrayTraits:
    """
    Array traits for Zserio structure, choice, union and enum types.
    """

    HAS_BITSIZEOF_CONSTANT = False

    def __init__(self, objectCreator: typing.Callable[[BitStreamReader, int], typing.Any]) -> None:
        """
        Constructor.

        :param objectCreator: Creator which creates object from the element index.
        """

        self._objectCreator = objectCreator # type: typing.Callable[[BitStreamReader, int], typing.Any]

    @staticmethod
    def bitSizeOf(bitPosition: int, value: typing.Any) -> int:
        """
        Returns length of Zserio object type stored in the bit stream in bits.

        :param bitPosition: Current bit position in bit stream.
        :param value: Zserio object type value.
        :returns: Length of given Zserio object type in bits.
        """

        return value.bitSizeOf(bitPosition)

    @staticmethod
    def initializeOffsets(bitPosition: int, value: typing.Any) -> int:
        """
        Initializes indexed offsets for the Zserio object type.

        :param bitPosition: Current bit stream position.
        :param value: Zserio object type value.
        :returns: Updated bit stream position which points to the first bit after the Zserio object type.
        """

        return value.initializeOffsets(bitPosition)

    def read(self, reader: BitStreamReader, index: int) -> typing.Any:
        """
        Reads Zserio object type from the bit stream.

        :param reader: Bit stream from which to read.
        :param index: Element index in the array.
        """

        return self._objectCreator(reader, index)

    @staticmethod
    def write(writer: BitStreamWriter, value: typing.Any) -> None:
        """
        Writes Zserio object type to the bit stream.

        :param writer: Bit stream where to write.
        :param value: Zserio object type to write.
        """

        value.write(writer)



import os
import subprocess
import sys
import shutil


class JavaNotFoundException(Exception):
    """
    This exception is raised if the Java executable is not found
    either through $JAVA_HOME or $PATH.
    """
    def __init__(self):
        super(self).__init__("Java was not found (checked $JAVA_HOME and $PATH).")


zs_jar_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "zserio.jar")
java_path = None


def find_java():
    global java_path
    if java_path:
        return java_path
    java_path = os.getenv("JAVA_HOME", None)
    if java_path:
        java_path = os.path.join(java_path, "bin/java")
        print(f"Using Java at {java_path} (found through JAVA_HOME).")
    else:
        java_path = shutil.which("java")
        if java_path:
            print(f"Using Java at {java_path} (found through PATH).")
        else:
            raise JavaNotFoundException()
    return java_path


def generate(src_file: str = "", module_name: str = "") -> None:
    """
    Description:

        Translate zserio code to python, and add the generated package
        to pythonpath. The generated sources will be placed under
        <src_file.zs>/../.zs-python-package[/module_name]. This path
        will be added to `sys.path`.

        The `module_name` argument allows to set a top-level python module name,
        under which the generated sources will be placed.

    Possible exceptions:

        `zserio.JavaNotFoundException`: Java was not found,
          therefore zserio.jar cannot be called.

        `subprocess.CalledProcessError`: Raised if
          `java -jar zserio.jar ...` produced a non-zero returncode.

    Examples:

        With top-level package:
            ```
            import zserio
            zserio.generate("myfile.zs", "mypackage")
            from mypackage.myfile import *
            ```

        Without top-level package:
            ```
            import zserio
            zserio.generate("myfile.zs")
            from myfile import *
            ```

    :param src_file: Source zserio file.
    :param module_name: (Optional) Top-level package directory name.
    :return: True if succesfull, False otherwise.
    """
    global zs_jar_path
    zs_pkg_path = os.path.dirname(os.path.abspath(src_file))
    zs_build_path = os.path.join(zs_pkg_path, ".zs-python-package")
    subprocess.run([
        find_java(), "-jar", zs_jar_path,
        "-src", zs_pkg_path,
        "-python", zs_build_path,
        *(("-setTopLevelPackage", module_name) if module_name else tuple()),
        os.path.basename(src_file)],
        check=True)
    if zs_build_path not in sys.path:
        sys.path.append(zs_build_path)



"""
The module provides help methods for bit position calculation.
"""

from zserio.exception import PythonRuntimeException

def alignTo(alignmentValue: int, bitPosition: int) -> int:
    """
    Aligns the bit size to the given alignment value.

    :param alignmentValue: Value to align.
    :param bitPosition: Current bit position where to apply alignment.
    :returns: Aligned bit position.
    """

    if bitPosition <= 0 or alignmentValue == 0:
        return bitPosition

    return (((bitPosition - 1) // alignmentValue) + 1) * alignmentValue

def bitsToBytes(numBits: int) -> int:
    """
    Converts number of bits to bytes.

    :param numBits: The number of bits to convert.
    :returns: Number of bytes
    :raises PythonRuntimeException: If number of bits to convert is not divisible by 8.
    """

    if numBits % 8 != 0:
        raise PythonRuntimeException("bitPosition: %d is not a multiple of 8!" % numBits)

    return numBits // 8

def bytesToBits(numBytes: int) -> int:
    """
    Converts number of bytes to bits.

    :param numBytes: The n number of bytes to convert.
    :returns: Number of bits.
    """

    return numBytes * 8



"""
The module implements exceptions for Zserio python runtime library.
"""

class PythonRuntimeException(Exception):
    """
    Exception thrown in case of an error in Zserio python runtime library.
    """



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

required_url = []
required = []
with open("requirements.txt", "r") as freq:
    for line in freq.read().split():
        if "://" in line:
            required_url.append(line)
        else:
            required.append(line)

with open("current/zserio/version.txt", "r") as version_file:
    version = version_file.read().strip()

packages = setuptools.find_packages("current")

setuptools.setup(
    name="zserio",
    version=version,
    url="http://zserio.org",
    author="Navigation Data Standard e.V.",
    author_email="support@nds-association.org",

    description="Zserio runtime and Python package generator.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    package_dir={'': 'current'},
    packages=packages,
    include_package_data=True,
    package_data={
        'zserio': ['zserio.jar']
    },

    install_requires=required,
    dependency_links=required_url,
    python_requires='>=3.6',

    license = "BSD-3 Clause",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License"
     ],
)



"""
Zserio Python runtime library

.. data:: VERSION_STRING

   Zserio Python runtime library version string.
"""
from .gen import *


from zserio.bitbuffer import BitBuffer
from zserio.bitreader import BitStreamReader
from zserio.bitwriter import BitStreamWriter
from zserio.exception import PythonRuntimeException
from zserio.service import ServiceInterface, ServiceException
from zserio.pubsub import PubsubInterface, PubsubException

import zserio.array
import zserio.bitfield
import zserio.bitposition
import zserio.bitsizeof
import zserio.builtin
import zserio.float
import zserio.hashcode
import zserio.limits

VERSION_STRING = "2.1.0-pre2"
